import asyncio
from typing import Union

import discord
from cpl_core.configuration.configuration_abc import ConfigurationABC
from cpl_core.database.context.database_context_abc import DatabaseContextABC
from cpl_core.logging import LoggerABC
from cpl_query.extension import List
from discord.ext.commands import Context

from gismo_core.abc.bot_service_abc import BotServiceABC
from gismo_core.abc.message_service_abc import MessageServiceABC
from gismo_core.configuration.server_settings import ServerSettings
from gismo_data.abc.client_repository_abc import ClientRepositoryABC


class MessageService(MessageServiceABC):

    def __init__(self, config: ConfigurationABC, logger: LoggerABC, bot: BotServiceABC, clients: ClientRepositoryABC, db: DatabaseContextABC):
        self._config = config
        self._logger = logger
        self._bot = bot
        self._clients = clients
        self._db = db
    
    async def delete_messages(self, messages: List[discord.Message]): 
        self._logger.debug(__name__, f'Try to delete {messages.count()} messages')
        for message in messages:
            server_st: ServerSettings = self._config.get_configuration(f'DSERVER_{message.guild.id}')
            await asyncio.sleep(server_st.message_delete_timer)
            await self.delete_message(message, mass_delete=True)
        self._logger.debug(__name__, 'Deleting messages finished')
        
    async def delete_message(self, message: discord.Message, mass_delete=False):
        server_st: ServerSettings = self._config.get_configuration(f'DSERVER_{message.guild.id}')
        if not mass_delete:
            await asyncio.sleep(server_st.message_delete_timer)
        self._logger.debug(__name__, f'Try to delete message:\n\t{message}\n\t{message.content}')
        guild_id = message.guild.id
        try:
            await message.delete()
        except Exception as e:
            self._logger.error(__name__, f'Deleting message failed', e)
        else:
            self._clients.append_deleted_message_count(self._bot.user.id, guild_id, 1)
            self._db.save_changes()
            self._logger.info(__name__, f'Deleted message {message}')
    
    async def send_channel_message(self, channel: discord.TextChannel, message: str):
        self._logger.debug(__name__, f'Try to send message\n\t{message}\n\tto: {channel}')
        msg = None
        try:
            msg = await channel.send(message)
        except Exception as e:
            self._logger.error(__name__, f'Send message to channel {channel.id} failed', e)
        else:
            self._logger.info(__name__, f'Sent message to channel {channel.id}')
            self._clients.append_sent_message_count(self._bot.user.id, channel.guild.id, 1)
            self._db.save_changes()
            await self.delete_message(msg)
    
    async def send_dm_message(self, message: str, receiver: Union[discord.User, discord.Member]):
        self._logger.debug(__name__, f'Try to send message\n\t{message}\n\tto: {receiver}')
        try:
            await receiver.send(message)
        except Exception as e:
            self._logger.error(__name__, f'Send message to user {receiver.id} failed', e)
        else:
            self._clients.append_sent_message_count(self._bot.user.id, receiver.guild.id, 1)
            self._db.save_changes()
            self._logger.info(__name__, f'Sent message to user {receiver.id}')
    
    async def send_ctx_msg(self, ctx: Context, message: str, file: discord.File = None):
        if ctx is None:
            self._logger.warn(__name__, 'Message context is empty')
            self._logger.debug(__name__, f'Message: {message}')
            return
        
        self._logger.debug(__name__, f'Try to send message\t\t{message}\n\tto: {ctx.channel}')
        msg = None
        try:
            if isinstance(message, discord.File):
                msg = await ctx.send(file=message)
            else:
                msg = await ctx.send(message)
        except Exception as e:
            self._logger.error(__name__, f'Send message to channel {ctx.channel.id} failed', e)
        else:
            self._logger.info(__name__, f'Sent message to channel {ctx.channel.id}')
            self._clients.append_sent_message_count(self._bot.user.id, ctx.guild.id, 1)
            self._db.save_changes()
            await self.delete_messages(List(discord.Message, [msg, ctx.message]))
